//*****************************************************************
//
// CMD_LINE.CPP: NDIR command line parsing
//
// Written by:	 Daniel D. Miller  (the Derelict)
//		 Jason Hood
//
// Last update:  01/16/02 16:18
//		 27 Aug, 2014
//
//*****************************************************************

#include "ndata.hpp"
#include "conio32.hpp"
#include <io.h>

#ifndef _O_U8TEXT
#define _O_U8TEXT 0x40000
#endif

// from NDIR32.CPP
extern void insert_target_filespec( const wchar_t* fstr );
// from FILELIST.CPP
extern const wchar_t* fprint[6][2];

static int low_ascii;

//*************************************************************
// Convert command-line switches into flags
//*************************************************************
static int update_switches( const wchar_t* argstr )
{
  int	   slen = 1;
  wchar_t* end;
  int	   num;

  switch (*argstr++)
  {
    case 'a':
      DWORD a, *ap;
      bool  attr, fnd, eq;
      a = 0;
      ap = &attrib_have;
      fnd = eq = false;
      attr = true;
      do
      {
	switch (towlower( *argstr++ ))
	{
	  case 'a': a |= FILE_ATTRIBUTE_ARCHIVE;   break;
	  case 'd': a |= FILE_ATTRIBUTE_DIRECTORY; break;
	  case 's': a |= FILE_ATTRIBUTE_SYSTEM;    break;
	  case 'h': a |= FILE_ATTRIBUTE_HIDDEN;    break;
	  case 'r': a |= FILE_ATTRIBUTE_READONLY;  break;

	  case '+':
	    if (a)
	    {
	      *ap = a;
	      a   = 0;
	      fnd = true;
	    }
	    ap = &attrib_must;
	  break;

	  case '-':
	    if (a)
	    {
	      *ap = a;
	      a   = 0;
	      fnd = true;
	    }
	    ap = &attrib_excl;
	  break;

	  case '=':
	    if (slen != 1)
	    {
	      --slen;
	      attr = false;
	    }
	    else
	    {
	      eq = true;
	    }
	  break;

	  case '?':
	    n.help  = 1;
	    helpptr = attrtxt;
	    attr    = false;
	  break;

	  default:
	    if (eq)
	    {
	      attrib_must = a;
	      attrib_excl = ~a;
	      fnd = true;
	    }
	    else if (a)
	    {
	      *ap = a;
	      fnd = true;
	    }
	    --slen;
	    attr = false;
	  break;
	}
	++slen;
      } while (attr);

      if (fnd)
	n.show_all  = 2;
      else
	n.show_all ^= 1;
    break;

    case 'b': n.batch ^= 1; break;

    case 'c':
      num = (int)wcstoul( argstr, &end, 10 );
      if (num > 9)
      {
	if (num > TEMPSTR_COUNT-2)
	  num = TEMPSTR_COUNT-2;
	reportcols = num;
	slen = end - argstr + 1;
      }
      else
      {
	n.clear ^= 1;
      }
    break;

    case 'd':
      if (*argstr == '2' || *argstr == '3')
      {
	n.tree = *argstr - '0';
	slen   = 2;
      }
      else
      {
	n.tree = 1;
	if (*argstr == '1')
	  slen = 2;
      }
    break;

    case 'e': n.sort = SORT_EXT;    break;

    case 'f': n.files_only    ^= 1; break;
    case 'g': n.dir_first     ^= 1; break;
    case 'h': n.horz          ^= 1; break;
    case 'i': n.drive_summary ^= 1; break;
    case 'j': n.low_ascii     ^= 1;
	      low_ascii       ^= 1; break;

    case 'k':
      if (*argstr == '\0' || *argstr == '/')
      {
	n.color ^= 1;
      }
      else
      {
	num = (uchar)wcstoul( argstr, &end, 0 );
	if (num == 0)
	{
	  while (*argstr && *argstr != '/' && slen < 5)
	  {
	    num = (num << 8) | (uchar)towlower( *argstr++ );
	    ++slen;
	  }
	}
	else
	{
	  slen = end - argstr + 1;
	}
	if (ccount < MAX_TARGETCLR)
	  targetclr[ccount++] = num;
      }
    break;

    case 'l': n.lfn_off ^= 1; break;

    case 'm':
      num = (int)wcstol( argstr, &end, 10 );
      if (num != 0 || *argstr == '0')
      {
	max_cols = num;
	slen = end - argstr + 1;
      }
      else
      {
	n.minimize ^= 1;
      }
    break;

    case 'n': n.sort = SORT_NAME; break;

    case 'o':
      if (*argstr == '0')
      {
	n.fdate_option = FDATE_LAST_WRITE;
	slen = 2;
      }
      else if (*argstr == '1')
      {
	n.fdate_option = FDATE_LAST_ACCESS;
	slen = 2;
      }
      else
      {
	n.fdate_option = FDATE_CREATE_TIME;
	if (*argstr == '2')
	  slen = 2;
      }
    break;

    case 'p': n.pause ^= 1; break;
    case 'q': n.horz  ^= 2; break;

    case 'r':
      num = (int)wcstol( argstr, &end, 10 );
      if (num != 0 || *argstr == '0')
      {
	reportrows = num;
	slen = end - argstr + 1;
      }
      else
      {
	n.reverse ^= 1;
      }
    break;

    case 's': n.sort = SORT_SIZE;  break;
    case 't': n.sort = SORT_DATE;  break;

    case 'u': n.ucase        ^= 1; break;

    case 'v':
      filesmax = (int)wcstol( argstr, &end, 10 );
      slen = end - argstr + 1;
    break;

    case 'w': n.showSHRfiles ^= 1; break;
    case 'x': n.exec_only    ^= 1; break;

    case 'y': n.numeric      ^= 1; break;
    case 'z': n.sort = SORT_NONE;  break;

    case '1': n.format = 0;        break;
    case '2': n.format = 1;        break;
    case '3': n.format = 2;        break;
    case '4': n.format = 3;        break;
    case '5': n.format = 4;        break;
    case '6': n.format = 5;        break;

    case ',': n.tree_short ^= 1;   break;
    case '.': n.recurse ^= 1;      break;
    case '$': n.hex ^= 1;          break;

    case '#':
      if (argstr[0] == '?' && argstr[1] == '\0')
      {
	n.help	= 1;
	helpptr = fmttxt;
	slen += wcslen( argstr );
      }
      else if (wcspbrk( argstr, L"# " ) == NULL)
      {
	batchfmt = NULL;
	find_ini_format( argstr );
	if (batchfmt == NULL)
	{
	  if ((argstr[1] == '\0' || argstr[1] == '/') &&
	      *argstr >= '1' && *argstr <= '6')
	    batchfmt = fprint[*argstr - '1'][!n.lfn_off];
	  else
	    error_exit( INV_FORMAT, argstr );
	}
	end = wcschr( (wchar_t*)argstr, '/' );
	if (end == NULL)
	  slen += wcslen( argstr );
	else
	  slen = end - argstr + 1;
      }
      else
      {
	batchfmt = argstr;
	slen += wcslen( argstr );
      }
      n.format = 6;
    break;

    case '?':
      n.help  = 1;
      helpptr = (*argstr == '?') ? ++slen, wildtxt : helptxt;
    break;
  }

  return slen;
}

//**********************************************************
void parse_command_string( const wchar_t* cmdstr )
{
  switch (*cmdstr)
  {
    case '-':
      if (wcscmp( cmdstr, L"--help" ) == 0)
      {
	n.help = 1;
	helpptr = helptxt;
	break;
      }
      else if (wcscmp( cmdstr, L"--version" ) == 0)
      {
	n.help = 1;
	helpptr = vertxt;
	break;
      }
    case '/':
      ++cmdstr; 		// skip the switch char
      while (*cmdstr)
	cmdstr += update_switches( cmdstr );
    break;

    default:
      insert_target_filespec( cmdstr );
    break;
  }
}

//**********************************************************
void parse_command_args( int startIdx, int argc, wchar_t** argv )
{
  wchar_t *argvptr, *spcptr;

  // see if we have an environment variable to deal with
  if (startIdx == 0)
  {
    argvptr = argv[0];
    while (1)
    {
      spcptr = wcschr( argvptr, ' ' );
      if (spcptr == 0)
      {
	parse_command_string( argvptr );
	break;
      }

      // fall thru here if space is found
      *spcptr++ = '\0';	// NULL-term first string, point to next string
      parse_command_string( argvptr );
      // -# will treat remainder as its format
      if (argvptr[1] == '#' && (argvptr[0] == '/' || argvptr[0] == '-'))
	break;
      argvptr = spcptr;
    }
    startIdx = 1;
  }

  // deal with normal command-line arguments
  for (; startIdx < argc; ++startIdx)
    parse_command_string( argv[startIdx] );
}

/**********************************************************************/
void verify_flags( void )
{
  cols	= get_window_cols() - 1;
  lines = get_window_rows() - 1;

  output = OUT_STD;
  if (!n.batch && !n.help)
  {
    if (redirected)
    {
      n.pause = 0;
      if (n.color)
	output = (redirected == 2) ? OUT_ANSI : OUT_HTML;
    }
    else if (n.color)
    {
      output = (hStdOut) ? OUT_SCREEN : OUT_ANSI;
    }
  }

  if (output == OUT_HTML)
  {
    _setmode( 1, _O_U8TEXT );
  }
  else if (!hStdOut)
  {
    wchar_t* lang = _wgetenv( L"LANG" );
    if ((lang && wcsstr( lang, L"UTF-8" )) || GetConsoleOutputCP() == CP_UTF8)
      _setmode( 1, _O_U8TEXT );
    else
      n.low_ascii = !low_ascii;
  }
  else if (output != OUT_SCREEN)
  {
    restore_console_attribs();
  }

  /************************************************/
  /*  Test for, and correct, inconsistent flags.  */
  /************************************************/

  // Check attribute options.
  if (n.show_all & 2)
    n.files_only = 0;

  // Set 'dir tree' conditions.
  if (n.tree || n.drive_summary)
  {
    n.minimize = 0;
    n.show_all |= 1;
  }

  // XTdir-style listing requires extension sorting.
  if (n.horz & 2)
    n.sort = SORT_EXT;

  //*************************************
  // Initialize line-drawing variables
  //*************************************
  if (n.low_ascii)
  {
    tline = bline = xline = iline = crosschar;
    vline = altvlinechar;
    dline = altdvlchar;
    hline = althlinechar;
  }
  else
  {
    tline = topchar;
    bline = bottomchar;
    vline = vlinechar;
    xline = xlinechar;
    dline = dvlinechar;
    hline = hlinechar;
    iline = ilinechar;
  }
}
