//*****************************************************************
//
// FILELIST.CPP: NDIR file-listing handlers
//
// Written by:   Daniel D. Miller  (the Derelict)
//		 Jason Hood
//
// Last update:  01/15/01 21:20AM
//		 19 Aug, 2013
//
//*****************************************************************

#include "ndata.hpp"

// from CONIO32.CPP
extern int redirected;

//*****************************************************************
const wchar_t* fprint[6][2] =
{
  //	       8.3			  LFN
  { L"#14S #a #d #M, #Y #T #F", L"#14S #a #d #M, #Y #T #F" },     // 1 column
  { L"#f  #7s  #d-#M-#y #t",    L"#6s #d-#M-#y #t #f"      },     // 2 columns
  { L"#f #6s #m-#y",            L"#6s #m-#y #f"            },     // 3 columns
  { L"#f #6s",                  L"#6s #f"                  },     // 4 columns
  { L"#f #l",                   L"#l #f"                   },     // 4 columns
  { L"#f",                      L"#f"                      },     // 6 columns
};

static const wchar_t* fmt;

//***************  function prototypes  ***************
static void filehead( void );
static void fileend( void );
void filesummary( void );

int total_filecount, total_dirscount;
static __int64 total_dbytes, total_dsbytes;

//*********************************************************
static void display_batch_mode( void )
{
  ffdata* ftemp;

  // find largest size
  maxlen[3] = 0;
  __int64 fsize = 0;
  for (ftemp = ftop; ftemp; ftemp = ftemp->next)
    if (ftemp->fsize > fsize)
      fsize = ftemp->fsize;
  int base = (n.hex) ? 16 : 10;
  do
  {
    ++maxlen[3];
    fsize /= base;
  } while (fsize);

  // initialise current time, to determine age
  format( NULL, L"", 0 );

  for (ftemp = ftop; ftemp; ftemp = ftemp->next)
  {
    format( ftemp, batchfmt, -1 );
    putwchar( '\n' );
  }
}

//*****************************************************************
static int line_len;		// width of column

static void get_columns( void )
{
  int	  curlen;
  __int64 fsize = 0;

  // find length of longest name and extension, largest size
  ZeroMemory( maxlen, sizeof(maxlen) );
  maxdir = 0;
  for (ffdata* ftemp = ftop; ftemp; ftemp = ftemp->next)
  {
    curlen = wcslen( ftemp->filename );
    if (curlen > maxlen[0])
      maxlen[0] = curlen;

    if (ftemp->dirflag && curlen > maxdir)
      maxdir = curlen;

    curlen = wcslen( ftemp->name );
    if (curlen > maxlen[1])
      maxlen[1] = curlen;

    curlen = wcslen( ftemp->ext );
    if (curlen > maxlen[2])
      maxlen[2] = curlen;

    if (ftemp->fsize > fsize)
      fsize = ftemp->fsize;
  }
  int base = (n.hex) ? 16 : 10;
  do
  {
    ++maxlen[3];
    fsize /= base;
  } while (fsize);

  line_len = format( NULL, fmt, 0 );

  // shortcut solution for very long filenames
  if (line_len + 1 + line_len > cols)
  {
    disp_cols = 1;
  }
  else
  {
    // compute line count:
    // (disp_cols * line_len) + (disp_cols - 1) * 1 < cols
    // (disp_cols * line_len) + disp_cols - 1 < cols
    // disp_cols * (line_len + 1) < cols + 1
    disp_cols = (cols + 1) / (line_len + 1);
  }
  if (max_cols && disp_cols > max_cols)
    disp_cols = max_cols;
  if (disp_cols > filecount + !n.minimize)
    disp_cols = filecount + !n.minimize;
}

//*****************************************************************
static void list_files_horizontally( void )
{
  int j = 0;

  // see how many columns we can support with requested format
  get_columns();

  filehead();
  for (ffdata* ftemp = ftop; ftemp; ++filesdisp, ftemp = ftemp->next)
  {
    format( ftemp, fmt, ++j );
    if (j == disp_cols)
    {
      ncrlf();
      j = 0;
    }
    else
    {
      nputc( n.colorframe, vline );
    }
  }
  if (j != 0)
  {
    // add in the remaining separators
    while (++j < disp_cols)
    {
      nput_char( n.colordefalt, ' ', line_len );
      nputc( n.colorframe, vline );
    }
    // put in closing newline
    ncrlf();
  }
  fileend();
}

//*****************************************************************
static void list_files_qwise( void )
{
  ffdata* ftemp;
  int	  col = 0, slen;
  int	  first_line = 1, new_line;
  const wchar_t* prev_ext = L"";
  int	  maxext;

  // first, find the longest extension in the current file list
  maxext = 0;
  for (ftemp = ftop; ftemp; ftemp = ftemp->next)
  {
    slen = wcslen( ftemp->ext );
    if (maxext < slen)
      maxext = slen;
  }

  // then list the files
  filehead();
  for (ftemp = ftop; ftemp; ++filesdisp, ftemp = ftemp->next)
  {
    // see if file extension is changing
    if (first_line || _wcsicmp( prev_ext, ftemp->ext ) != 0)
    {
      if (first_line)
	first_line = 0;
      else
	ncrlf();
      nputf( ftemp->color, L"%-*s: ", maxext, ftemp->ext );
      col      = maxext + 2;
      new_line = 1;
      prev_ext = ftemp->ext;
    }

    // see if next filename is going to overrun line;
    // if so, start next line...
    slen = (lfn_supported) ? wcslen( ftemp->name ) + 2 : 9;
    if (col + slen > cols + 1)
    {
      if (lfn_supported)
	nputc( ftemp->color, ',' );
      ncrlf();
      nput_char( ftemp->color, ' ', maxext + 2 );
      col      = maxext + 2;
      new_line = 1;
    }
    // if not starting new line, add comma separator
    if (new_line)
      new_line = 0;
    else if (lfn_supported)
      nputs( ftemp->color, L", " );
    // select appropriate color and print the filename
    nputf( ftemp->color, (lfn_supported) ? L"%s" : L"%-8s ", ftemp->name );
    col += slen;
  }
  ncrlf();
  fileend();
}

//*****************************************************************
static void list_files_vertically( void )
{
  int	  rows = 0, partrows, fcount;
  int	  j, k, r;
  int	  pcols[42];		// these arrays must have enough entries for
  ffdata* ftemps[42];		// the maximum number of columns on the screen
  wchar_t buf[TEMPSTR_COUNT];

  get_columns();
  if (disp_cols > 42)
    disp_cols = 42;

  wmemset( buf, hline, cols );
  buf[cols]   = '\n';
  buf[cols+1] = '\0';
  for (j = line_len, k = disp_cols; --k > 0; j += line_len + 1)
    buf[j] = iline;

  filehead();

  if (disp_cols == 1)
  {
    r = filecount + linecnt;
  }
  else if (redirected)
  {
    r = (reportrows == 0) ? filecount + linecnt :
	(reportrows == -1 || reportrows == 1) ? lines : abs( reportrows );
  }
  else
  {
    r = (reportrows == 0) ? filecount + linecnt : lines;
  }

  filesdisp = j = 0;
  fcount    = filecount;
  ftemps[0] = ftop;
  while (1)
  {
    if (rows == 0)
    {
      if (fcount > (r - linecnt) * disp_cols)
      {
	rows	 = r - linecnt;
	partrows = 0;
      }
      else
      {
	rows	 = fcount / disp_cols;
	partrows = fcount % disp_cols;
      }

      for (j = 0; j < disp_cols; j++) pcols[j] = rows;
      for (j = 0; j < partrows;  j++) pcols[j]++;

      if (partrows > 0)
	rows++;

      //************************************************
      // split the file list into (columns) lists.
      //************************************************
      for (j = 1; j < disp_cols; j++)
      {
	ftemps[j] = ftemps[j-1];
	for (k = pcols[j-1]; k > 0; k--)
	  ftemps[j] = ftemps[j]->next;
      }
      j = 0;
    }

    if (filesdisp < filecount)
    {
      format( ftemps[j], fmt, j+1 );
      ftemps[j] = ftemps[j]->next;
      ++filesdisp;
      --fcount;
    }
    // if no files left to display, fill in row with spaces
    else
    {
      nput_char( n.colordefalt, ' ', line_len );
    }

    if (++j == disp_cols)
    {
      ncrlf();
      if (fcount == 0)
	break;
      if (--rows == 0)
      {
	ftemps[0] = ftemps[disp_cols-1];
	linecnt   = 0;				// remove the overlap line
	if (reportrows > 0)			// add a "page" separator
	  nputs( n.colorframe, buf );
      }
      j = 0;
    }
    else
    {
      nputc( n.colorframe, vline );
    }
  } // loop forever
  fileend();
}

/****************************************************************/
/*  Display directory data in "files" mode.                     */
/****************************************************************/
void display_files( void )
{
  if (n.batch)
  {
    display_batch_mode();
    return;
  }

  //****************************************************
  // normal display mode
  //****************************************************
  if (filecount == 0)
  {
    filehead();
    nputs( n.colordefalt, L"No matching files found.\n" );
    fileend();
    return;
  }

  //************************************************
  // present normal file listings
  //************************************************
  fmt = (n.format == 6) ? batchfmt : fprint[n.format][lfn_supported];
  if (n.horz & 2)
    list_files_qwise();
  else if (n.horz & 1)
    list_files_horizontally();
  else
    list_files_vertically();
}

/*****************************************************************/
static void filehead( void )
{
  int j, k;

  j = base_len + 1 + 16 + wcslen( volume_name );

  //****************************************************
  // If minimize is selected, display short header/footer
  //****************************************************
  if (n.minimize)
  {
    nputs( n.colornhead, base_path );

    if (j <= cols && (!n.recurse || total_filecount == 0))
    {
      nput_char( n.colornhead, ' ', cols - j + 1 );
      nputs( n.colornhead, L"Volume label is " );
      nputs( n.colorxhead, volume_name );
    }
    ncrlf();
  }
  //****************************************************
  // If minimize is not selected, display full header/footer
  //****************************************************
  else
  {
    nputs( n.colornhead, L"Directory of " );
    nputs( n.colornhead, base_path );
    j += 13;
    if (j > cols)
      ncrlf();
    else
      nput_char( n.colornhead, ' ', cols - j + 1 );

    if (!n.recurse || total_filecount == 0)
    {
      nputs( n.colornhead, L"Volume label is " );
      nputs( n.colorxhead, volume_name );
    }
    ncrlf();

    wmemset( tempstr, dline, cols );
    tempstr[cols] = '\0';
    if (filecount)
    {
      k = (filecount < disp_cols) ? filecount + 1 : disp_cols;
      for (j = line_len; --k > 0; j += line_len + 1)
	tempstr[j] = tline;
    }
    nputs( n.colorframe, tempstr );
    ncrlf();
  }

  filesdisp = 1;
}

/*****************************************************************/
void put_disk_summary( void )
{
  nputs( n.colornhead, L"Total disk space: " );
  nputs( n.colorxhead, i64toss( diskbytes ) );
  nputs( n.colornhead, L" bytes; Free space: " );
  nputs( n.colorxhead, i64toss( diskfree ) );
  nputs( n.colornhead, L" bytes\n" );
}

/*****************************************************************/
static void fileend( void )
{
  int	  j, k;
  __int64 dbytes, dsbytes;
  int	  wid;

  // get directory info
  dbytes = dsbytes = 0;
  for (ffdata* ftemp = ftop; ftemp; ftemp = ftemp->next)
  {
    dbytes  += ftemp->fsize;
    wcscpy( base_path + base_len, ftemp->filename );
    dsbytes += get_disk_size( base_path, ftemp->fsize, ftemp->compflag );
  }
  if (n.recurse)
  {
    total_dbytes += dbytes;
    total_dsbytes += dsbytes;
    total_filecount += filecount;
    total_dirscount += dirscount;
  }

  wid = 0;
  //**************************************************************
  // If minimize is selected, display short header/footer
  //**************************************************************
  if (n.minimize)
  {
    if (dirscount != 0)
    {
      wid  = nputf( n.colorxhead, L"%d", dirscount );
      wid += nputf( n.colornhead, L" dir%s, ", L"s" + (dirscount == 1) );
    }
    wid += nputf( n.colorxhead, L"%d", filecount - dirscount );
    wid += nputf( n.colornhead, L" file%s: ", L"s" + (filecount-dirscount == 1) );
    wid += nputs( n.colorxhead, i64toss( dbytes ) );
    wid += nputs( n.colornhead, L" <" );
    wid += nputs( n.colorxhead, i64toss( dsbytes ) );
    if (!n.recurse)
    {
      int meg = 0;
      wchar_t* df = i64toss( diskfree );
      wchar_t* db = i64toss( diskbytes );
      if (wid + 9 + (int)wcslen( df ) + 2 + (int)wcslen( db ) + 1 > cols)
      {
	// If free space is >= 10GB, use megabytes.  (Due to the use of static
	// buffers in i64toss, convert both at the same time, so one doesn't
	// overwrite the other.  Ugh, that's why you don't use static buffers!)
	if (diskfree >= 10000000000)
	{
	  df = i64toss( diskfree / 1000000 );
	  db = i64toss( diskbytes / 1000000 );
	  meg = 1;
	}
      }
      if (wid + 9 + (int)wcslen( df ) + meg + 2
	  + (int)wcslen( db ) + meg + 1 <= cols)
	nputs( n.colornhead, L">; " );
      else
      {
	nputc( n.colornhead, '>' );
	ncrlf();
      }
      nputs( n.colornhead, L"Disk: " );
      nputs( n.colorxhead, df );
      if (meg)
	nputc( n.colorxhead ^ 8, 'M' );
      nputs( n.colornhead, L" <" );
      nputs( n.colorxhead, db );
      if (meg)
	nputc( n.colorxhead ^ 8, 'M' );
    }
    nputc( n.colornhead, '>' );
    ncrlf();
  }
  //**************************************************************
  // If minimize is not selected, display full header/footer
  //**************************************************************
  else
  {
    // draw the bottom framing line
    wmemset( tempstr, dline, cols );
    tempstr[cols] = '\0';
    if (filecount)
    {
      k = (filecount < disp_cols) ? filecount + 1 : disp_cols;
      for (j = line_len; --k > 0; j += line_len + 1)
	tempstr[j] = bline;
    }
    nputs( n.colorframe, tempstr );
    ncrlf();

    // draw the ending labels
    if (dirscount != 0)
    {
      wid  = nputf( n.colorxhead, L"%d", dirscount );
      wid += nputf( n.colornhead, L" dir%s, ", L"s" + (dirscount == 1) );
    }
    wid += nputf( n.colorxhead, L"%d", filecount - dirscount );
    wid += nputs( n.colornhead, (filecount - dirscount == 1)
				? L" file of " : L" files total " );
    wid += nputs( n.colorxhead, i64toss( dbytes ) );
    wid += nputf( n.colornhead, L" byte%s,", L"s" + (dbytes == 1) );
    wchar_t* ds = i64toss( dsbytes );
    if (wid + 1 + 6 + (int)wcslen( ds ) + 21 <= cols)
      nputc( n.colornhead, ' ' );
    else
      ncrlf();
    nputf( n.colornhead, L"using " );
    nputs( n.colorxhead, ds );
    nputs( n.colornhead, L" bytes of disk space." );
    ncrlf();

    if (!n.recurse)
      put_disk_summary();
  }
}

/*****************************************************************/
void filesummary( void )
{
  if (total_filecount == 0)
  {
    // let this handle the no files found message
    int recurse = n.recurse;
    n.recurse = false;
    display_files();
    n.recurse = recurse;
    return;
  }

  ncrlf();

  int wid = 0;
  //**************************************************************
  // If minimize is selected, display short header/footer
  //**************************************************************
  if (n.minimize)
  {
    if (total_dirscount != 0)
    {
      wid  = nputf( n.colorxhead, L"%d", total_dirscount );
      wid += nputf( n.colornhead, L" dir%s, ", L"s" + (total_dirscount == 1) );
    }
    wid += nputf( n.colorxhead, L"%d", total_filecount - total_dirscount );
    wid += nputf( n.colornhead, L" file%s: ",
			       L"s" + (total_filecount - total_dirscount == 1) );
    wid += nputs( n.colorxhead, i64toss( total_dbytes ) );
    wid += nputs( n.colornhead, L" <" );
    wid += nputs( n.colorxhead, i64toss( total_dsbytes ) );
    wchar_t* df = i64toss( diskfree );
    wchar_t* db = i64toss( diskbytes );
    if (wid + 9 + (int)wcslen( df ) + 2 + (int)wcslen( db ) + 1 <= cols)
      nputs( n.colornhead, L">; " );
    else
    {
      nputc( n.colornhead, '>' );
      ncrlf();
    }
    nputs( n.colornhead, L"Disk: " );
    nputs( n.colorxhead, df );
    nputs( n.colornhead, L" <" );
    nputs( n.colorxhead, db );
    nputs( n.colornhead, L">\n" );
  }
  //**************************************************************
  // If minimize is not selected, display full header/footer
  //**************************************************************
  else
  {
    // draw the ending labels
    if (total_dirscount != 0)
    {
      wid  = nputf( n.colorxhead, L"%d", total_dirscount );
      wid += nputf( n.colornhead, L" dir%s, ", L"s" + (total_dirscount == 1) );
    }
    wid += nputf( n.colorxhead, L"%d", total_filecount - total_dirscount );
    wid += nputs( n.colornhead, (total_filecount - total_dirscount == 1)
				? L" file of " : L" files total " );
    wid += nputs( n.colorxhead, i64toss( total_dbytes ) );
    wid += nputf( n.colornhead, L" byte%s,", L"s" + (total_dbytes == 1) );
    wchar_t* ds = i64toss( total_dsbytes );
    if (wid + 1 + 6 + (int)wcslen( ds ) + 21 <= cols)
      nputc( n.colornhead, ' ' );
    else
      ncrlf();
    nputs( n.colornhead, L"using " );
    nputs( n.colorxhead, ds );
    nputs( n.colornhead, L" bytes of disk space." );
    ncrlf();

    put_disk_summary();
  }
}
