//*****************************************************************************
//
// I64TOSTR.CPP: Convert unsigned 64-bit integer to thousands-separated string.
//
// Written by:	 Jason Hood
//
// Last update:  16 Dec, 2005
//
//*****************************************************************************

#include "ndata.hpp"

static wchar_t sep;

static wchar_t* convert( unsigned long num, wchar_t* str, int& pos )
{
  do
  {
    *--str = '0' + (wchar_t)(num % 10);
    if (++pos == 3)
    {
      *--str = sep;
      pos = 0;
    }
    num /= 10;
  } while (num);

  return str;
}

//*****************************************************************************
wchar_t* i64toss( unsigned __int64 uli )
{
  int	   pos = 0;
  wchar_t* strptr;
  wchar_t* base;
  static wchar_t buf[2][32];	// rotate between two buffers
  static int	 idx;

  if (sep == '\0')
  {
    if (GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_STHOUSAND, &sep, 1 ) == 0)
      sep = ',';
  }

  strptr = buf[idx];
  idx	^= 1;

  if (n.hex)
  {
    int len = _snwprintf( strptr, 32, L"%I64X", uli );
    int size = 5;
    while (len > 4)
    {
      wmemmove( strptr + len - 3, strptr + len - 4, size );
      strptr[len - 4] = sep;
      size += 5;
      len -= 4;
    }
  }
  else
  {
    strptr += 31;
    if (uli > (__int64)999999999)
    {
      base   = strptr - 12;
      strptr = convert( (unsigned long)(uli % 1000000000), strptr, pos );
      uli   /= (__int64)1000000000;
      while (strptr > base)
      {
	*--strptr = '0';
	if (++pos == 3)
	{
	  *--strptr = sep;
	  pos = 0;
	}
      }
    }
    strptr = convert( (unsigned long)uli, strptr, pos );
    if (*strptr == sep)
      ++strptr;
  }
  return strptr;
}
