//*****************************************************************
//
// NDATA.CPP:	NDIR global data declarations
//
// Written by:  Daniel D. Miller  (the Derelict)
//		Jason Hood
//
// Last update: 01/15/01 20:25
//		19 Aug, 2013
//
//*****************************************************************

#include "ndata.hpp"

//*****************************************************************
// CONFIGURATION VARIABLES
//*****************************************************************
ndir_data n =
{
  // default colors
  0x02, // nhead
  0x0A, // xhead
  0x08, // frame
  0x01, // attr
  0x05, // size
  0x0C, // date
  0x04, // time
  0x09, // dir
  0x18, // SHR
  0x06, // defalt

  //*********************************************************
  // program state flags
  //*********************************************************
  0, // uppercase = false
  0, // sort by filename
  0, // no reverse
  1, // numeric
  0, // decimal file sizes
  0, // clearscreen = false
  1, // pause on fullscreen = true
  3, // display format = name & size
  1, // Minimize display (make abbreviated header/footer)
  0, // display normal files, not tree
  0, // 0=normal tree, 1=show only one level of children
  0, // recurse subdirectories = false
  1, // list directories first = true
  0, // show all files = false
  0, // show only files = false
  0, // show executables only = false
  0, // list files horizontally = false
  0, // use low-ASCII characters = false
  0, // display in batch mode = false
  1, // display files in color = true
  1, // showSHRfiles in color = true
  0, // display disk_drive summary
  0, // file-date option (MSDOS 7.00+ only)
  0, // 1 = force old DOS 8.3 filenames in MSDOS 7.00+
  0, // display help screen = false
};

attrib_list *attr_table[2*28], *attr_tail[2*28];

uchar dtree_colors[MAX_DIR_ENTRY] =
{
  3, 4, 5, 6, 7, 8, 9, 10
};
int max_color_level = 8;

//*****************************************************************
// END OF CONFIGURATION VARIABLES
//*****************************************************************

int lfn_supported = 1; // always true for 32-bit version

WIN32_FIND_DATA fdata; // long-filename file struct

__int64 diskbytes, diskfree, clbytes;
wchar_t	volume_name[MAXPATH];
bool	is_ntfs;

//*****************************************************************
ffdata *ftop, *ftail;

wchar_t** target;
wchar_t** target_path;
int	  tcount;	// number of target filespecs
int	  tcapacity;	// number of allocated filespecs
DWORD	  attrib_have = ~0, attrib_must, attrib_excl;
DWORD	  targetclr[MAX_TARGETCLR];
int	  ccount;

// name of drive+path without filenames
wchar_t   base_path[MAXPATH];
int	  base_len;	// length of base_path
int	  wild_pos;	// position of wildcard within target

int start, finish;
int filecount, dirscount;
int filesdisp;		// number of files displayed
int disp_cols;		// calculated screen columns for listings
int max_cols;		// maximum screen columns
int maxlen[4];		// longest length of each component
int maxdir;
int filesmax;		// maximum number of files to display

const wchar_t* batchfmt = L"#f";

wchar_t tempstr[TEMPSTR_COUNT];
wchar_t tline, bline, vline, xline, dline, hline, iline;

int cols, lines, linecnt;
int reportcols = 80;		// line length for redirected output
int reportrows =  1;		// number of rows for vertical file list
int output;

//*********************************************************
// NDIR help screen
//*********************************************************
const wchar_t* const helptxt[] =
{
  L"jNDIR by Jason Hood <jadoxa@yahoo.com.au> and Daniel D. Miller (the Derelict).",
  L"Version " PVER L" (" PDATE L").  Freeware.",
  L"http://jndir.adoxa.vze.com/",
  L"",
  L"USAGE:  jNDIR <filespecs> -options or /options",
  L"",
  L" OPTIONS                      FUNCTIONS",
  L"   -a *     List all files (include hidden and system).",
  L"   -aATTR   Match attributes (use -a? for help).",
  L"   -c *     Clear screen before listing.",
  L"   -cN      Specify columns for redirected output (10 to 778).",
  L"   -p *     Pause on full screen.",
  L"   -m *     Minimize size of header and footer.",
  L"   -w *     Use special colors for system/hidden/readonly files.",
  L"   -d[1]    Dir Tree: normal size display (work with -s, -n (default), -r).",
  L"   -d2      Dir Tree: file/directory counts",
  L"   -d3      Dir Tree: mixed size and file/directory counts.",
  L"   -, *     Dir Tree: show only one level of subdirectories.",
  L"   -. *     Recurse subdirectories.",
  L"   -n       Sort by name.",
  L"   -e        \"   by extension.",
  L"   -s        \"   by file size, smallest first.",
  L"   -t        \"   by date, oldest first.",
  L"   -z        \"   by directory order (no sort).",
  L"   -r *     Reverse normal sort order.",
  L"   -y *     Toggle numeric name sorting.",
  L"   -rN      Specify rows for vertical file listing (0=maximum, 1=screen,",
  L"               negative will remove the dividing line).",
  L"   -vN      Maximum number of files to display (negative taken from the end).",
  L"   -1       Display one column,   with name/size/date/attr.",
  L"   -2          \"    two   \"   ,   with name/size/date.",
  L"   -3          \"    three \"   ,   with name/size/mmYY.",
  L"   -4          \"    four  \"   ,   with name/size.",
  L"   -5          \"    four  \"   ,   with name/age.",
  L"   -6          \"    six   \"   ,   with name only.",
  L"   -mN      Maximum number of columns to display (0 for whatever fits).",
  L"   -#FMT    Specify the format to display (use -#? for help).",
  L"               If FMT doesn't contain hash or space, it's read from the config.",
  L"   -oN      Date/Time display: 0=Last Write, 1=Last Access, 2=File Created.",
  L"   -$       Display sizes using hexadecimal.",
  L"   -i *     Display drive summary for all drives in system.",
  L"   -l *     Toggle long filename flag.",
  L"   -k *     Toggle color mode.",
  L"   -kID     Select files matching color or ID (see NDIR.INI).",
  L"   -j *     Use standard ASCII (for redirection).",
  L"   -u *     List short filenames in UPPERCASE.",
  L"   -x *     List executables only (.EXE,.COM,.BAT,.CMD,.BTM).",
  L"   -g *     List directories first.",
  L"   -h *     List files horizontally.",
  L"   -q *     Group files by extension.",
  L"   -f *     List files only (no directories).",
  L"   -b *     Batch mode; one file per line, suitable for a batch file.",
  L"   -?       Display help screen.",
  L"   -??      Display filespec help screen.",
  L"",
  L"NOTE: items with a * after the flag are TOGGLES.",
  0
};

const wchar_t* const vertxt[] =
{
  L"jNDIR version " PVER L" (" PDATE L").",
  0
};

const wchar_t* const wildtxt[] =
{
  L"NDIR wildcard characters.",
  L"",
  L"*\tmatch zero or more characters",
  L"?\tmatch any one character",
  L"[SET]\tmatch any character in SET",
  L"[!SET]\tmatch any character not in SET",
  L"!p\tmatch everything except \"p\"",
  L"p;q!r;s\tmatch \"p\" or \"q\", but not \"r\" or \"s\"",
  L"",
  L"SET may contain a range of characters \"[a-z]\".  To include ']'",
  L"place it first; to include '-' place it first or after a range.",
  L"To match '!' or ';' place them in a set.",
  0
};

const wchar_t* const attrtxt[] =
{
  L"NDIR attribute matching.",
  L"",
  L"-a[HAVE][+MUST][-EXCL]",
  L"-a=[ATTR]",
  L"",
  L"HAVE\tattribute must include one of these",
  L"MUST\tattribute must have all of these",
  L"EXCL\tattribute must not have any of these",
  L"ATTR\tattribute must match exactly",
  L"",
  L"where attribute consists of:",
  L"  a\tarchive",
  L"  d\tdirectory",
  L"  s\tsystem",
  L"  h\thidden",
  L"  r\tread-only",
  L"",
  L"Examples: -ash:   match system and/or hidden (with anything else)",
  L"          -as+h:  match system AND hidden (with anything else)",
  L"          -a-sh:  match everything except system or hidden",
  L"          -ash-r: match system and/or hidden, but not with read-only",
  L"          -a=sh:  match exactly system and hidden (nothing else allowed)",
  0
};

const wchar_t* const fmttxt[] =
{
  L"NDIR format specifiers.",
  L"",
  L"#f\tfilename",
  L"#F\tfilename or [directory]",
  L"#n\tname",
  L"#e\textension (including dot)",
  L"#E\textension (without dot)",
  L"#p\tfilename with path",
  L"#P\tpath only, including trailing backslash",
  L"#q\tquoted filename",
  L"#Q\tquoted filename with path",
  L"",
  L"#Ns\tsize, without thousands separator",
  L"#NS\tsize, with thousands separator",
  L"\tN specifies the width to use (for decimal; hex will always use the",
  L"\tlongest); default is the longest size, four is the minimum.  Will",
  L"\tuse \"[DIR]\" (or \"[D]\") for a directory, except if #F is used,",
  L"\twhich will use spaces.",
  L"",
  L"#a\tattribute",
  L"",
  L"#d\tday (01-31)",
  L"#D\tday of week (Sun-Sat)",
  L"#m\tmonth (01-12)",
  L"#M\tmonth (Jan-Dec)",
  L"#y\ttwo-digit year",
  L"#Y\tfour-digit year",
  L"#o\tlocal short date",
  L"#O\tlocal long date",
  L"",
  L"#T\thh:mm:ss",
  L"#t\thh:mm",
  L"",
  L"#l\tage (units depend on how old it is)",
  L"#L\tage in days (or years if really old)",
  0
};

const wchar_t* const* helpptr;
