//************************************************************
//
// NDATA.HPP:	Header file for NDIR directory lister.
//
// Written by:  Daniel D. Miller
//		Jason Hood
//
// Last update: 05/31/02 16:57
//		19 Sep, 2014
//
//************************************************************

#define PVER  L"1.24"
#define PDATE L"12 January, 2015"

#define UNICODE
#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x500
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <wchar.h>

#ifndef _MAX_PATH
#define _MAX_PATH 260
#endif
#define MAXPATH (_MAX_PATH*2)

typedef unsigned char uchar;
typedef unsigned long ulong;

#define MAXLINE 260
#define MAX_TARGETCLR 20

//*********************************************************
struct attrib_list
{
  attrib_list* next;
  wchar_t* ext;
  DWORD id;
  uchar attr;

  attrib_list( const wchar_t*, uchar, DWORD );
};
extern attrib_list* attr_table[2*28], *attr_tail[2*28];
int which_attrib_table( const wchar_t* );

#define MAX_DIR_ENTRY 20
extern uchar dtree_colors[MAX_DIR_ENTRY];
extern int max_color_level;

//*********************************************************
union i64tol
{
  __int64 i;
  ulong   u[2];
};

//*****************************************************************
// CONFIGURATION VARIABLES
//*****************************************************************
struct ndir_data
{
  uchar colornhead;
  uchar colorxhead;
  uchar colorframe;
  uchar colorattr;
  uchar colorsize;
  uchar colordate;
  uchar colortime;
  uchar colordir;
  uchar colorSHR;
  uchar colordefalt;

  uchar ucase;
  uchar sort;		// 0=name, 1=ext, 2=size, 3=date/time, 4=none,
  uchar reverse;	// reverse normal sort order
  uchar numeric;	// numeric name sorting
  uchar hex;		// hexadecimal file sizes
  uchar clear;
  uchar pause;
  uchar format;
  uchar minimize;
  uchar tree;		// 0=files, 1=sizes, 2=counts, 3=counts & sizes
  uchar tree_short;	// 0=normal tree, 1=show only one level of children
  uchar recurse;	// recurse into subdirectories
  uchar dir_first;
  uchar show_all;
  uchar files_only;
  uchar exec_only;
  uchar horz;
  uchar low_ascii;
  uchar batch;
  uchar color;
  uchar showSHRfiles;
  uchar drive_summary;
  uchar fdate_option;	// 0 = LAST_MODIFY, 1 = FILE_CREATE (MSDOS 7.00+ only)
  uchar lfn_off;	// 0 = force DOS 8.3 filenames (MSDOS 7.00+ only)
  uchar help;
};
extern ndir_data n;

// fdate_option flags
#define FDATE_LAST_WRITE  0
#define FDATE_LAST_ACCESS 1
#define FDATE_CREATE_TIME 2

//*****************************************************************
// END OF CONFIGURATION VARIABLES
//*****************************************************************

//***************************************************************
extern int lfn_supported;	// always true for 32-bit version

extern WIN32_FIND_DATA fdata;

extern __int64 diskbytes, diskfree, clbytes;
extern wchar_t volume_name[MAXPATH];
extern bool    is_ntfs;

//************************************************************
#define topchar       0x2564	/*    */
#define bottomchar    0x2567	/*    */
#define vlinechar     0x2502	/*    */
#define xlinechar     0x256A	/*    */
#define dvlinechar    0x2550	/*    */
#define hlinechar     0x2500	/*    */
#define ilinechar     0x253C	/*    */
#define crosschar     0x2B	/*  +  */
#define altvlinechar  0x7C	/*  |  */
#define altdvlchar    0x3D	/*  =  */
#define althlinechar  0x2D	/*  -  */

extern wchar_t tline, bline, vline, xline, dline, hline, iline;

/************************************************************/
struct ffdata
{
  FILETIME ft;
  __int64  fsize;
  wchar_t* filename;
  wchar_t* name;
  wchar_t* ext;
  uchar    attrib;
  uchar    dirflag;
  uchar    compflag;
  uchar    color;
  ffdata*  next;
};
extern ffdata* ftop;
extern ffdata* ftail;

//**********************************************************
extern wchar_t** target;
extern wchar_t** target_path;
extern int	 tcount;
extern int	 tcapacity;
extern DWORD	 attrib_have, attrib_must, attrib_excl;
extern DWORD	 targetclr[MAX_TARGETCLR];
extern int	 ccount;

extern wchar_t	 base_path[MAXPATH];	// name of drive+path without filenames
extern int	 base_len;		// length of base_path
extern int	 wild_pos;		// position of wildcard within target

//*********************************************************
// NSORT.CPP
//*********************************************************
enum sort_modes
{
  SORT_NAME, SORT_EXT, SORT_SIZE, SORT_DATE, SORT_NONE
};

/************************************************************/
/*  error codes and messages.                               */
/************************************************************/
enum error_values { DATA_OKAY = 0, OUT_OF_MEMORY, INV_DRIVE, INV_FORMAT };

__declspec(noreturn)
void error_exit( int index, const wchar_t* outstr );

extern const wchar_t Version[];

//************************************************************
// ndata.cpp
//************************************************************
extern const wchar_t* const  helptxt[];
extern const wchar_t* const  vertxt[];
extern const wchar_t* const  wildtxt[];
extern const wchar_t* const  attrtxt[];
extern const wchar_t* const  fmttxt[];
extern const wchar_t* const* helpptr;

#define TEMPSTR_COUNT (MAXLINE+MAXPATH)
extern wchar_t tempstr[TEMPSTR_COUNT];
extern int     filecount;	// number of files and directories found
extern int     dirscount;	// number of directories found
extern int     filesdisp;	// number of files displayed
extern int     disp_cols;	// number of columns to use
extern int     max_cols;	// but no more than this
extern int     maxlen[4];	// longest filename/name/extension/size
extern int     maxdir;		// longest directory
extern int     filesmax;	// maximum number of files to display

extern const wchar_t* batchfmt;	// custom or batch mode format

extern int lines, cols;
extern int linecnt;		// display line counter
extern int reportcols, reportrows; // number of columns/rows for file output
extern int output;		// the type of output

enum { OUT_STD, OUT_SCREEN, OUT_HTML, OUT_ANSI };

extern int start, finish;

//*****************  function prototypes  **********************

extern "C"
{
#include "filmatch.h"
}

// CMD_LINE.CPP
void verify_flags( void );

// FILELIST.CPP
void display_files( void );

// FILEREAD.CPP
void file_listing( void );
bool find_first( const wchar_t* path );
bool find_next( void );

// I64TOSTR.CPP
wchar_t* i64toss( unsigned __int64 num );

// TREELIST.CPP
void tree_listing( void );

// NDISPLAY.CPP
void info( void );
void display_logo( void );
int  format( const ffdata* fptr, const wchar_t* fmtstr, int col );
void ncrlf( void );
void nputc( uchar attr, wchar_t outchr );
int  nputs( uchar attr, const wchar_t* outstr );
int  nputf( uchar attr, const wchar_t* fmt, ... );
void nput_char( uchar attr, wchar_t chr, int count );
void nput_line( uchar attr, wchar_t chr );

// DISKPARM.CPP
void display_drive_summary( void );
void get_disk_info( const wchar_t* dstr );
__int64 get_disk_size( const wchar_t* name, __int64 fsize, BOOL compressed );

// JNDIR.CPP
void find_ini_format( const wchar_t* fmt );
