//*****************************************************************
//
// NDISPLAY.CPP: NDIR display handlers
//
// Written by:	 Daniel D. Miller  (the Derelict)
//		 Jason Hood
//
// Last update:  01/15/01 21:20
//		 18 Dec, 2005
//
//*****************************************************************

#include <stdarg.h>
#include "ndata.hpp"
#include "conio32.hpp"


// from i64tostr.cpp
extern wchar_t sep;

static const wchar_t attrchr[] = L"-adshr";

/*
static const wchar_t monthstr[13][4] =
{
  L"???", L"Jan", L"Feb", L"Mar", L"Apr", L"May", L"Jun",
	  L"Jul", L"Aug", L"Sep", L"Oct", L"Nov", L"Dec"
};

static const wchar_t daystr[7][4] =
{
  L"Sun", L"Mon", L"Tue", L"Wed", L"Thu", L"Fri", L"Sat"
};
*/

static void form_ansi_attr( uchar attr );


static bool  dirsize;
static uchar attr;
static ULONGLONG  now_secs;
static SYSTEMTIME now_st;

//*********************************************************
// This does not take into account wrapped lines.
//*********************************************************
static void testpause( void )
{
  if (!n.pause || redirected)
    return;

  if (++linecnt >= lines)
  {
    nputc( n.colornhead, '[' );
    nputf( n.colorxhead, L"%d", filesdisp );
    nputc( n.colornhead, '/' );
    nputf( n.colorxhead, L"%d", filecount );
    nputs( n.colornhead, L"] Press " );
    if (!hStdOut)
    {
      nputs( n.colornhead, L"Enter to continue or ESC to exit " );
      form_ansi_attr( 0 );
      fflush( stdout );
      if (getchar() == 27)
	error_exit( DATA_OKAY, NULL );
      fflush( stdin );
      fputws( L"\033[A\033[K", stdout );
    }
    else
    {
      nputs( n.colornhead, L"any key to continue or ESC to exit " );
      if (get_scode() == 27)
	error_exit( DATA_OKAY, NULL );

      dreturn();
      dclreol();
    }
    linecnt = 1;
  }
}

//**************************************************
void info( void )
{
  for (int j = 0; helpptr[j]; ++j)
    _putws( helpptr[j] );
}

//**************************************************
// linecnt = number of lines written.
//**************************************************
void display_logo( void )
{
  if (n.batch)
    return;

  if (n.clear && !redirected)
  {
    if (!hStdOut)
      fputws( L"\033[2J", stdout );
    else
      dclrscr();
  }

  linecnt = 0;
}

//********************************************************
static void print_size( const ffdata* fptr, wchar_t fmt, int wid )
{
  wchar_t  buf[32];
  wchar_t* numstr;
  int	   len, size, suf, sufw;
  static const wchar_t suffix[] = L"KMG";

  if (fptr->dirflag)
  {
    if (dirsize)
      nputf( n.colordir, L"%*s", wid, (wid == 4) ? L"[D]" : L"[DIR]" );
    else
      nput_char( n.colorsize, ' ', wid );
    return;
  }

  if (fmt == 's')
  {
    _snwprintf( buf, 32, (n.hex) ? L"%I64X" : L"%I64u", fptr->fsize );
    numstr = buf;
    sufw   = 3;
  }
  else
  {
    numstr = i64toss( fptr->fsize );
    sufw   = 4;
  }

  len = size = wcslen( numstr );
  if (len <= wid || n.hex)
  {
    nputf( n.colorsize, L"%*s", wid, numstr );
    return;
  }
  suf = -1;
  do
  {
    len -= sufw;
    ++suf;
  } while (len >= wid);
  numstr[len] = '\0';
  nputf( n.colorsize, L"%*s", wid - 1, numstr );
  nputc( n.colorsize ^ 8, suffix[suf] );
}

//********************************************************
static void print_name( const ffdata* fptr, const wchar_t* fmt, int col )
{
  const wchar_t* fmtstr;

  if (lfn_supported)
  {
    col = (fmt[1] == '\0' && col == disp_cols) ? 1 : maxlen[0];
    if (fptr->dirflag && *fmt == 'F')
    {
      _snwprintf( tempstr, TEMPSTR_COUNT, L"[%s]%*s", fptr->filename,
		  (col == 1) ? 0 : col - 2 - wcslen( fptr->filename ), L"" );
    }
    else
    {
      _snwprintf( tempstr, TEMPSTR_COUNT, L"%-*s", col, fptr->filename );
    }
  }
  else
  {
    fmtstr = (*fmt == 'F') ? ((fptr->dirflag) ? L"[%-*s%-*s]" : L"%-*s%-*s  ")
			   : L"%-*s%-*s";
    _snwprintf( tempstr, TEMPSTR_COUNT, fmtstr,
		maxlen[1], fptr->name, maxlen[2], fptr->ext );
  }

  nputs( attr, tempstr );
}

//********************************************************
static void print_str( const wchar_t* str, int wid, const wchar_t* fmt, int col )
{
  if (col == -1)
    fputws( str, stdout );
  else if (col == disp_cols && fmt[1] == '\0')
    nputs( attr, str );
  else
    nputf( attr, L"%-*s", wid, str );
}

//********************************************************
void print_attr( const ffdata* fptr )
{
  wchar_t attrstr[8];
  int	  a = fptr->attrib;

  // detect non-standard file attributes and display specially.
  if (a > 0x3F)
  {
    nputf( n.colorattr ^ 8, L" 0x%2X", a );
  }
  else
  {
    attrstr[0] = (a & FILE_ATTRIBUTE_ARCHIVE)	? attrchr[1] : attrchr[0];
    attrstr[1] = (a & FILE_ATTRIBUTE_DIRECTORY) ? attrchr[2] : attrchr[0];
    attrstr[2] = (a & FILE_ATTRIBUTE_SYSTEM)	? attrchr[3] : attrchr[0];
    attrstr[3] = (a & FILE_ATTRIBUTE_HIDDEN)	? attrchr[4] : attrchr[0];
    attrstr[4] = (a & FILE_ATTRIBUTE_READONLY)	? attrchr[5] : attrchr[0];
    attrstr[5] = '\0';
    nputs( n.colorattr, attrstr );
  }
}

//********************************************************
static void print_age( const FILETIME* fft, const SYSTEMTIME* fst, bool days )
{
  SYSTEMTIME later_st, earlier_st;
  static wchar_t period[] = L"smhdy"; // secs, mins, hours, days, years
  unsigned t;
  int w;

  ULARGE_INTEGER uli;
  uli.LowPart = fft->dwLowDateTime;
  uli.HighPart = fft->dwHighDateTime;
  ULONGLONG secs = uli.QuadPart / 10000000;
  if (secs > now_secs)
  {
    secs -= now_secs;
    later_st   = *fst;
    earlier_st = now_st;
    _wcsupr( period );
  }
  else
  {
    secs = now_secs - secs;
    later_st   = now_st;
    earlier_st = *fst;
    _wcslwr( period );
  }
  if (days)
  {
    unsigned d = (unsigned)(secs / (24*60*60));
    if (iswupper( period[4] ) && d < 1000)
    {
      nputf( n.colordate, L"%+4d", d );
    }
    else if (iswlower( period[4] ) && d < 10000)
    {
      nputf( n.colordate, L"%4u", d );
    }
    else
    {
      int years = later_st.wYear - earlier_st.wYear;
      if (later_st.wMonth < earlier_st.wMonth)
	--years;
      nputf( n.colordate, L"%3d", years );
      nputc( n.colordate ^ 8, period[4] );
    }
  }
  else if (secs < 60)
  {
    // under a minute, use seconds
    nputf( n.colortime, L"%3u", (unsigned)secs );
    nputc( n.colortime ^ 8, period[0] );
  }
  else if (secs < 10*60)
  {
    // under ten minutes, use minutes & seconds
    t = (unsigned)secs % 60;
    w = (t == 0) ? 3 : 1;
    nputf( n.colortime, L"%*u", w, (unsigned)secs / 60 );
    nputc( n.colortime ^ 8, period[1] );
    if (t != 0)
      nputf( n.colortime, L"%.2u", t );
  }
  else if (secs < 60*60)
  {
    // under one hour, use minutes
    nputf( n.colortime, L"%3u", (unsigned)secs / 60 );
    nputc( n.colortime ^ 8, period[1] );
  }
  else if (secs < 10*60*60)
  {
    // under ten hours, use hours & minutes
    t = (unsigned)secs % (60*60) / 60;
    w = (t == 0) ? 3 : 1;
    nputf( n.colortime, L"%*u", w, (unsigned)secs / (60*60) );
    nputc( n.colortime ^ 8, period[2] );
    if (t != 0)
      nputf( n.colortime, L"%.2u", t );
  }
  else if (secs < 24*60*60)
  {
    // under a day, use hours
    nputf( n.colortime, L"%3u", (unsigned)secs / (60*60) );
    nputc( n.colortime ^ 8, period[2] );
  }
  else if (secs < 10*24*60*60)
  {
    // under ten days, use days & hours
    t = (unsigned)secs % (24*60*60) / (60*60);
    w = (t == 0) ? 3 : 1;
    nputf( n.colordate, L"%*u", w, (unsigned)secs / (24*60*60) );
    nputc( n.colordate ^ 8, period[3] );
    if (t != 0)
      nputf( n.colortime, L"%.2u", t );
  }
  else if (secs < 366*24*60*60)
  {
    // under a year, use days
    nputf( n.colordate, L"%3u", (unsigned)secs / (24*60*60) );
    nputc( n.colordate ^ 8, period[3] );
  }
  else
  {
    // over a year, use years and months
    int years = later_st.wYear - earlier_st.wYear;
    int months = later_st.wMonth - earlier_st.wMonth;
    if (months < 0)
    {
      months += 12;
      --years;
    }
    if (years >= 10 || months == 0)
    {
      nputf( n.colordate, L"%3d", years );
      nputc( n.colordate ^ 8, period[4] );
    }
    else
    {
      nputf( n.colordate, L"%d", years );
      nputc( n.colordate ^ 8, period[4] );
      nputf( n.colordate, L"%.2d", months );
    }
  }
}

/******************************************************************/
int format( const ffdata* fptr, const wchar_t* fmt, int col )
{
  FILETIME   ft;
  SYSTEMTIME st;
  static int datewid;
  static wchar_t timesep = ':';

  if (fptr != NULL)
  {
    attr = (fptr->attrib & 7);
    attr = (n.showSHRfiles && attr) ? n.colorSHR | attr : fptr->color;

    FileTimeToLocalFileTime( &fptr->ft, &ft );
    FileTimeToSystemTime( &ft, &st );
  }
  else
  {
    GetLocalTime( &now_st );
    SystemTimeToFileTime( &now_st, &ft );
    ULARGE_INTEGER uli;
    uli.LowPart = ft.dwLowDateTime;
    uli.HighPart = ft.dwHighDateTime;
    now_secs = uli.QuadPart / 10000000;
  }

  int len = 0, wid, num;
  for (; *fmt; ++fmt)
  {
    if (*fmt != '#')
    {
      if (fptr == NULL)
	++len;
      // use current color for literals
      else if (output == OUT_SCREEN)
	dputc( *fmt );
      else
	putwchar( *fmt );
    }
    else
    {
      wid = 0;
      while (iswdigit( *++fmt ))
	wid = wid * 10 + *fmt - '0';

      switch (*fmt)
      {
	case 'f':
	case 'F':
	  if (fptr == NULL)
	  {
	    dirsize = true;
	    if (*fmt == 'F')
	    {
	      dirsize = false;
	      if (maxdir + 2 > maxlen[0])
		maxlen[0] = maxdir + 2;
	    }
	    len += maxlen[0];
	  }
	  else if (col == -1)
	  {
	    // no space padding in batch mode
	    wprintf( (*fmt == 'f' || !fptr->dirflag) ? L"%s%s%s" : L"[%s%s%s]",
		     target_path[start], base_path + wild_pos, fptr->filename );
	  }
	  else
	  {
	    print_name( fptr, fmt, col );
	  }
	break;

	case 'n':
	  if (fptr == NULL)
	    len += maxlen[1];
	  else
	    print_str( fptr->name, maxlen[1], fmt, col );
	break;

	case 'e':
	  if (fptr == NULL)
	    len += maxlen[2];
	  else
	    print_str( fptr->ext, maxlen[2], fmt, col );
	break;

	case 'E':
	  wid = (maxlen[2] == 0) ? 0 : maxlen[2] - 1;
	  if (fptr == NULL)
	    len += wid;
	  else
	    print_str( (*fptr->ext == '.') ? fptr->ext+1 : L"", wid, fmt, col );
	break;

	case 'p':
	  if (fptr == NULL)
	    len += base_len + maxlen[0];
	  else
	  {
	    _snwprintf( tempstr, TEMPSTR_COUNT, L"%s%s", base_path, fptr->filename );
	    print_str( tempstr, base_len + maxlen[0], fmt, col );
	  }
	break;

	case 'P':
	  if (fptr == NULL)
	    len += base_len;
	  else
	    print_str( base_path, base_len, fmt, col );
	break;

	case 'q':
	  if (fptr == NULL)
	    len += maxlen[0] + 2;
	  else
	  {
	    _snwprintf( tempstr, TEMPSTR_COUNT, L"\"%s%s%s\"",
		     target_path[start], base_path+wild_pos, fptr->filename );
	    print_str( tempstr, maxlen[0] + 2, fmt, col );
	  }
	break;

	case 'Q':
	  if (fptr == NULL)
	    len += base_len + maxlen[0] + 2;
	  else
	  {
	    _snwprintf( tempstr, TEMPSTR_COUNT, L"\"%s%s\"", base_path, fptr->filename );
	    print_str( tempstr, base_len + maxlen[0] + 2, fmt, col );
	  }
	break;

	case 's':
	case 'S':
	  if (wid == 0 || n.hex)
	  {
	    wid = maxlen[3];
	    if (*fmt == 'S')
	      wid += (wid - 1) / (n.hex ? 4 : 3);
	  }
	  if (wid < 4)
	    wid = 4;
	  if (fptr == NULL)
	    len += wid;
	  else
	    print_size( fptr, *fmt, wid );
	break;

	case 'a':
	  if (fptr == NULL)
	    len += 5;
	  else
	    print_attr( fptr );
	break;

	case 't':
	  if (fptr == NULL)
	  {
	    GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_STIME, &timesep, 1 );
	    len += 5;
	  }
	  else
	  {
	    nputf( n.colortime, L"%02d%c%02d", st.wHour, timesep, st.wMinute );
	  }
	break;

	case 'T':
	  if (fptr == NULL)
	  {
	    GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_STIME, &timesep, 1 );
	    len += 8;
	  }
	  else
	  {
	    nputf( n.colortime, L"%02d%c%02d%c%02d", st.wHour, timesep,
				st.wMinute, timesep, st.wSecond );
	  }
	break;

	case 'D':
	  if (fptr == NULL)
	    len += 3;
	  else
	  {
	    //nputs( n.colordate, daystr[st.wDayOfWeek] );
	    GetDateFormat( LOCALE_USER_DEFAULT, 0, &st, L"ddd", tempstr, 4 );
	    nputs( n.colordate, tempstr );
	  }
	break;

	case 'd':
	  if (fptr == NULL)
	    len += 2;
	  else
	    nputf( n.colordate, L"%02d", st.wDay );
	break;

	case 'm':
	  if (fptr == NULL)
	    len += 2;
	  else
	    nputf( n.colordate, L"%02d", st.wMonth );
	break;

	case 'M':
	  if (fptr == NULL)
	    len += 3;
	  else
	  {
	    //nputs( n.colordate, monthstr[st.wMonth] );
	    GetDateFormat( LOCALE_USER_DEFAULT, 0, &st, L"MMM", tempstr, 4 );
	    nputs( n.colordate, tempstr );
	  }
	break;

	case 'y':
	  if (fptr == NULL)
	    len += 2;
	  else
	    nputf( n.colordate, L"%02d", st.wYear % 100 );
	break;

	case 'Y':
	  if (fptr == NULL)
	    len += 4;
	  else
	    nputf( n.colordate, L"%04d", st.wYear );
	break;

	case 'o':
	case 'O':
	  if (fptr == NULL)
	  {
	    if (wid == 0)
	    {
	      DWORD flag = (*fmt == 'o') ? DATE_SHORTDATE : DATE_LONGDATE;
	      for (ffdata* ftemp = ftop; ftemp; ftemp = ftemp->next)
	      {
		FileTimeToLocalFileTime( &ftemp->ft, &ft );
		FileTimeToSystemTime( &ft, &st );
		num = GetDateFormat( LOCALE_USER_DEFAULT, flag,
				     &st, NULL, NULL, 0 ) - 1;
		if (num > wid)
		  wid = num;
	      }
	    }
	    datewid = wid;
	    len += wid;
	  }
	  else
	  {
	    num = GetDateFormat( LOCALE_USER_DEFAULT,
				 (*fmt == 'o') ? DATE_SHORTDATE : DATE_LONGDATE,
				 &st, NULL, tempstr, MAXPATH ) - 1;
	    if (datewid > num)
	      nput_char( n.colordate, ' ', datewid - num );
	    nputs( n.colordate, tempstr );
	  }
	break;

	case 'L':
	case 'l':
	  if (fptr == NULL)
	    len += 4;
	  else
	    print_age( &ft, &st, *fmt == 'L' );
	break;

	default:
	  if (fptr == NULL)
	    ++len;
	  else
	  {
	    if (output == OUT_SCREEN)
	      dputc( *fmt );
	    else
	      putwchar( *fmt );
	  }
	break;
      }
    }
  }

  return len;
}

/******************************************************************/
static const ulong html_dos_attr[16] =
{
  0x000000, 0x000080, 0x008000, 0x008080,
  0x800000, 0x800080, 0x808000, 0xC0C0C0,
  0x808080, 0x0000FF, 0x00FF00, 0x00FFFF,
  0xFF0000, 0xFF00FF, 0xFFFF00, 0xFFFFFF
};

static bool  inspan;
static uchar oldattr;

static void form_html_attr( uchar attr )
{
  ulong fg, bg;

  if (attr == oldattr)
    return;
  oldattr = attr;

  if (inspan)
  {
    fputws( L"</span>", stdout );
    inspan = false;
  }

  if (attr == n.colordefalt)
    return;

  fg = html_dos_attr[attr & 0xF];
  bg = html_dos_attr[attr >> 4];
  if ((attr & 0xF) == (n.colordefalt & 0xF))
    wprintf( L"<span style=\"background-color:#%06lX\">", bg );
  else if ((attr >> 4) == (n.colordefalt >> 4))
    wprintf( L"<span style=\"color:#%06lX\">", fg );
  else
    wprintf( L"<span style=\"color:#%06lX;background-color:#%06lX\">", fg, bg );

  inspan = true;
}

/******************************************************************/
static void form_ansi_attr( uchar attr )
{
  static int fg = -1, bg = -1, bold, underline;
  static const int attr2ansi[] = { 0, 4, 2, 6, 1, 5, 3, 7 };
  int	  f, b, B, u;
  wchar_t seq[32];
  int	  len;

  if (attr == 0)
  {
    // Restore original color before doing a new line.
    fputws( L"\033[m", stdout );
    oldattr = 0;
    fg = bg = -1;
    return;
  }

  if (attr == oldattr)
    return;
  oldattr = attr;

  f = attr2ansi[attr & 7];
  b = attr2ansi[(attr >> 4) & 7];
  B = attr & 8;
  u = attr & 0x80;

  // Make no assumptions about the default color.
  if (fg == -1)
  {
    bold = !B;
    underline = !u;
  }

  len = 0;
  seq[len++] = '\33';
  seq[len++] = '[';
  if (B && !bold)
  {
    seq[len++] = '1';
  }
  else if (!B && bold)
  {
    seq[len++] = '2';
    seq[len++] = '2';
  }
  if (f != fg)
  {
    if (B != bold)
      seq[len++] = ';';
    seq[len++] = '3';
    seq[len++] = '0' + f;
  }
  if (u && !underline)
  {
    if (len != 2)
      seq[len++] = ';';
    seq[len++] = '4';
  }
  else if (!u && underline)
  {
    if (len != 2)
      seq[len++] = ';';
    seq[len++] = '2';
    seq[len++] = '4';
  }
  if (b != bg)
  {
    if (u != underline)
      seq[len++] = ';';
    seq[len++] = '4';
    seq[len++] = '0' + b;
  }
  seq[len++] = 'm';
  seq[len++] = '\0';

  fputws( seq, stdout );

  fg = f;
  bg = b;
  bold = B;
  underline = u;
}

/******************************************************************/
void output_html_header( const wchar_t* title )
{
  if (output != OUT_HTML)
    return;

  wprintf( L"<html><head>\n"
	   L"<meta HTTP-EQUIV=\"content-type\" CONTENT=\"text/html; charset=UTF-8\">\n"
	   L"<title>%s</title></head>\n"
	   L"<body text=#%06lX bgcolor=#%06lX>\n"
	   L"<pre>\n",
	   title, html_dos_attr[n.colordefalt & 0xF],
		  html_dos_attr[n.colordefalt >> 4] );
  oldattr = n.colordefalt;
}

/******************************************************************/
void output_html_footer( void )
{
  if (output == OUT_HTML)
  {
    if (inspan)
      _putws( L"</span>" );
    _putws( L"</pre>\n</body></html>" );
  }
  else if (output == OUT_ANSI)
    fputws( L"\033[m", stdout );
}

/******************************************************************/
const wchar_t* gethtmlc( wchar_t outchr )
{
  const wchar_t* str;
  static wchar_t chr[2] = L"?";

  switch (outchr)
  {
    case '&': str = L"&amp;"; break;
    case '<': str = L"&lt;";  break;
    case '>': str = L"&gt;";  break;
    default:
      *chr = outchr;
      str  = chr;
    break;
  }

  return str;
}

/******************************************************************/
void ncrlf( void )
{
  switch (output)
  {
    case OUT_ANSI:
      form_ansi_attr( 0 );

    case OUT_STD:
    case OUT_HTML:
      putwchar( '\n' );
    break;

    case OUT_SCREEN:
      dnewline();
    break;
  }

  testpause();
}

/******************************************************************/
void nputc( uchar attr, wchar_t outchr )
{
  switch (output)
  {
    case OUT_ANSI:
      form_ansi_attr( attr );

    case OUT_STD:
      putwchar( outchr );
    break;

    case OUT_HTML:
      form_html_attr( attr );
      fputws( gethtmlc( outchr ), stdout );
    break;

    case OUT_SCREEN:
      set_text_attr( attr );
      dputc( outchr );
    break;
  }
}

/******************************************************************/
int nputs( uchar attr, const wchar_t* outstr )
{
  switch (output)
  {
    case OUT_ANSI:
      form_ansi_attr( attr );

    case OUT_STD:
      fputws( outstr, stdout );
    break;

    case OUT_HTML:
      form_html_attr( attr );
      while (*outstr)
	fputws( gethtmlc( *outstr++ ), stdout );
    break;

    case OUT_SCREEN:
      set_text_attr( attr );
      dputs( outstr );
    break;
  }
  return wcslen( outstr );
}

/******************************************************************/
int nputf( uchar attr, const wchar_t* fmt, ... )
{
  va_list arg;
  int len;

  va_start( arg, fmt );
  _vsnwprintf( tempstr, TEMPSTR_COUNT, fmt, arg );
  len = nputs( attr, tempstr );
  va_end( arg );
  return len;
}

/******************************************************************/
void nput_char( uchar attr, wchar_t chr, int count )
{
  const wchar_t* str;
  int j;

  switch (output)
  {
    case OUT_ANSI:
      form_ansi_attr( attr );

    case OUT_STD:
      for (j = count; j > 0; j--)
	putwchar( chr );
    break;

    case OUT_HTML:
      form_html_attr( attr );
      str = gethtmlc( chr );
      for (j = count; j > 0; j--)
	fputws( str, stdout );
    break;

    case OUT_SCREEN:
      dputnchar( chr, attr, count );
    break;
  }
}

/******************************************************************/
void nput_line( uchar attr, wchar_t chr )
{
  nput_char( attr, chr, cols );
  ncrlf();
}
